package org.kura.myasset.asset;

import static java.util.Objects.isNull;
import static java.util.Objects.nonNull;
import static java.util.Objects.requireNonNull;
import static org.eclipse.kura.channel.ChannelType.READ_WRITE;
import static org.eclipse.kura.channel.ChannelType.WRITE;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.eclipse.kura.KuraException;
import org.eclipse.kura.asset.AssetConfiguration;
import org.eclipse.kura.asset.provider.BaseAsset;
import org.eclipse.kura.channel.Channel;
import org.eclipse.kura.channel.ChannelFlag;
import org.eclipse.kura.channel.ChannelRecord;
import org.eclipse.kura.channel.ChannelStatus;
import org.eclipse.kura.channel.ChannelType;
import org.eclipse.kura.localization.LocalizationAdapter;
import org.eclipse.kura.localization.resources.WireMessages;
import org.eclipse.kura.type.TypedValue;
import org.eclipse.kura.type.TypedValues;
import org.eclipse.kura.util.collection.CollectionUtil;
import org.eclipse.kura.wire.WireEmitter;
import org.eclipse.kura.wire.WireEnvelope;
import org.eclipse.kura.wire.WireHelperService;
import org.eclipse.kura.wire.WireReceiver;
import org.eclipse.kura.wire.WireRecord;
import org.eclipse.kura.wire.WireSupport;
import org.osgi.service.component.ComponentContext;
import org.osgi.service.wireadmin.Wire;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public final class MyWireAsset extends BaseAsset implements WireEmitter, WireReceiver {

    private static final String ERROR_NOT_SPECIFIED_MESSAGE = "ERROR NOT SPECIFIED";

    private static final String ASSET_NAME = "assetName";

    private static final String TIMESTAMP = "timestamp";

    private static final String PROPERTY_SEPARATOR = "_";

    /** Configuration PID Property. */
    private static final String CONF_PID = "org.kura.myasset.asset.MyWireAsset";

    private static final Logger logger = LoggerFactory.getLogger(MyWireAsset.class);

    private static final WireMessages message = LocalizationAdapter.adapt(WireMessages.class);

    private volatile WireHelperService wireHelperService;

    private WireSupport wireSupport;

    /**
     * Binds the Wire Helper Service.
     *
     * @param wireHelperService
     *            the new Wire Helper Service
     */
    public void bindWireHelperService(final WireHelperService wireHelperService) {
    	logger.info("****************HELLO********************");
        if (isNull(this.wireHelperService)) {
            this.wireHelperService = wireHelperService;
        }
    }

    /**
     * Unbinds the Wire Helper Service.
     *
     * @param wireHelperService
     *            the new Wire Helper Service
     */
    public void unbindWireHelperService(final WireHelperService wireHelperService) {
    	logger.info("****************HELLO********************");
        if (this.wireHelperService == wireHelperService) {
            this.wireHelperService = null;
        }
    }

    /**
     * OSGi service component activation callback.
     *
     * @param componentContext
     *            the component context
     * @param properties
     *            the service properties
     */
    @Override
    protected void activate(final ComponentContext componentContext, final Map<String, Object> properties) {
    	logger.info("****************HELLO********************");
        logger.debug(message.activatingWireAsset());
        super.activate(componentContext, properties);
        this.wireSupport = this.wireHelperService.newWireSupport(this);
        logger.debug(message.activatingWireAssetDone());
    }

    /**
     * OSGi service component update callback.
     *
     * @param properties
     *            the service properties
     */
    @Override
    public void updated(final Map<String, Object> properties) {
    	logger.info("****************HELLO********************");
        logger.debug(message.updatingWireAsset());
        super.updated(properties);
        logger.debug(message.updatingWireAssetDone());
    }

    /**
     * OSGi service component deactivate callback.
     *
     * @param context
     *            the context
     */
    @Override
    protected void deactivate(final ComponentContext context) {
    	logger.info("****************HELLO********************");
        logger.debug(message.deactivatingWireAsset());
        super.deactivate(context);
        logger.debug(message.deactivatingWireAssetDone());
    }

    /** {@inheritDoc} */
    @Override
    public void consumersConnected(final Wire[] wires) {
    	logger.info("****************HELLO********************");
        this.wireSupport.consumersConnected(wires);
    }

    /** {@inheritDoc} */
    @Override
    protected String getFactoryPid() {
    	logger.info("****************HELLO********************");
        return CONF_PID;
    }

    /**
     * This method is triggered as soon as the wire component receives a Wire
     * Envelope. After it receives a {@link WireEnvelope}, it checks for all associated
     * channels to read and write and perform the operations accordingly. The
     * order of executions are performed the following way:
     *
     * <ul>
     * <li>Perform all read operations on associated reading channels</li>
     * <li>Perform all write operations on associated writing channels</li>
     * <ul>
     *
     * Both the aforementioned operations are performed as soon as this Wire Component
     * receives {@code Non Null} {@link WireEnvelop} from its upstream Wire Component(s).
     *
     * @param wireEnvelope
     *            the received {@link WireEnvelope}
     * @throws NullPointerException
     *             if {@link WireEnvelope} is null
     */
    @Override
    public void onWireReceive(final WireEnvelope wireEnvelope) {
    	logger.info("****************HELLO********************");
        requireNonNull(wireEnvelope, message.wireEnvelopeNonNull());
        logger.debug(message.wireEnvelopeReceived(), this.wireSupport);

        if (hasReadChannels()) {
            try {
                emitChannelRecords(readAllChannels());
            } catch (final KuraException e) {
                logger.error(message.errorPerformingRead(), e);
            }
        }

        final List<WireRecord> records = wireEnvelope.getRecords();
        for (WireRecord wireRecord : records) {
            final List<ChannelRecord> channelRecordsToWrite = determineWritingChannels(wireRecord);
            writeChannels(channelRecordsToWrite);
        }
    }

    /**
     * Determine the channels to write
     *
     * @param records
     *            the list of {@link WireRecord}s to parse
     * @return list of Channel Records containing the values to be written
     * @throws NullPointerException
     *             if argument is null
     */
    private List<ChannelRecord> determineWritingChannels(final WireRecord record) {
    	logger.info("****************HELLO********************");
        requireNonNull(record, message.wireRecordNonNull());

        final List<ChannelRecord> channelRecordsToWrite = CollectionUtil.newArrayList();
        final AssetConfiguration assetConfiguration = getAssetConfiguration();
        final Map<String, Channel> channels = assetConfiguration.getAssetChannels();
        for (final Entry<String, Channel> channelEntry : channels.entrySet()) {
            final Channel channel = channelEntry.getValue();
            final String channelName = channel.getName();
            final ChannelType channelType = channel.getType();

            if (channelType != WRITE && channelType != READ_WRITE) {
                continue;
            }

            Map<String, TypedValue<?>> wireRecordProperties = record.getProperties();

            if (wireRecordProperties.containsKey(channelName)) {
                final TypedValue<?> value = wireRecordProperties.get(channelName);
                if (channel.getValueType() == value.getType()) {
                    channelRecordsToWrite.add(channel.createWriteRecord(value));
                }
            }
        }
        return channelRecordsToWrite;
    }

    /**
     * Emit the provided list of channel records to the associated wires.
     *
     * @param channelRecords
     *            the list of channel records conforming to the aforementioned
     *            specification
     * @throws NullPointerException
     *             if provided records list is null
     * @throws IllegalArgumentException
     *             if provided records list is empty
     */
    private void emitChannelRecords(final List<ChannelRecord> channelRecords) {
    	logger.info("****************HELLO********************");
        requireNonNull(channelRecords, message.channelRecordsNonNull());
        if (channelRecords.isEmpty()) {
            throw new IllegalArgumentException(message.channelRecordsNonEmpty());
        }

        final Map<String, TypedValue<?>> wireRecordProperties = new HashMap<>();
        try {
            wireRecordProperties.put(ASSET_NAME, TypedValues.newStringValue(getKuraServicePid()));
        } catch (KuraException e) {
            logger.error(message.configurationNonNull(), e);
        }

        for (final ChannelRecord channelRecord : channelRecords) {
            final ChannelStatus channelStatus = channelRecord.getChannelStatus();
            final ChannelFlag channelFlag = channelStatus.getChannelFlag();
            final String channelName = channelRecord.getChannelName();

            final TypedValue<?> typedValue;
            if (channelFlag == ChannelFlag.FAILURE) {
                logErrorMessage(channelStatus);
                continue;
            } else {
                typedValue = channelRecord.getValue();
            }

            wireRecordProperties.put(channelName, typedValue);

            wireRecordProperties.put(channelName + PROPERTY_SEPARATOR + TIMESTAMP,
                    TypedValues.newLongValue(channelRecord.getTimestamp()));
        }
        wireRecordProperties.put("new Key", TypedValues.newStringValue("Hello"));
        final WireRecord wireRecord = new WireRecord(wireRecordProperties);
        this.wireSupport.emit(Arrays.asList(wireRecord));
    }

    private void logErrorMessage(final ChannelStatus channelStatus) {
    	logger.info("****************HELLO********************");
        String errorMessage = ERROR_NOT_SPECIFIED_MESSAGE;
        final Exception exception = channelStatus.getException();
        final String exceptionMsg = channelStatus.getExceptionMessage();
        if (nonNull(exception) && nonNull(exceptionMsg)) {
            errorMessage = exceptionMsg + " " + exception.toString();
        } else if (isNull(exception) && nonNull(exceptionMsg)) {
            errorMessage = exceptionMsg;
        } else if (nonNull(exception)) {
            errorMessage = exception.toString();
        }
        logger.error(errorMessage);
    }

    /**
     * Perform Channel Write operation
     *
     * @param channelRecordsToWrite
     *            the list of {@link ChannelRecord}s
     * @throws NullPointerException
     *             if the provided list is null
     */
    private void writeChannels(final List<ChannelRecord> channelRecordsToWrite) {
    	logger.info("****************HELLO********************");
        requireNonNull(channelRecordsToWrite, message.channelRecordsNonNull());
        if (channelRecordsToWrite.isEmpty()) {
            return;
        }

        try {
            write(channelRecordsToWrite);
        } catch (final KuraException e) {
            logger.error(message.errorPerformingWrite(), e);
        }
    }

    /** {@inheritDoc} */
    @Override
    public Object polled(final Wire wire) {
    	logger.info("****************HELLO********************");
        return this.wireSupport.polled(wire);
    }

    /** {@inheritDoc} */
    @Override
    public void producersConnected(final Wire[] wires) {
    	logger.info("****************HELLO********************");
        this.wireSupport.producersConnected(wires);
    }

    /** {@inheritDoc} */
    @Override
    public void updated(final Wire wire, final Object value) {
    	logger.info("****************HELLO********************");
        this.wireSupport.updated(wire, value);
    }
}